import math
import re
import sys
import time
from difflib import get_close_matches
from itertools import compress
from constants import NUMERICS
import numpy as np
import pandas as pd

class DataAnalyzer:
    
    def __init__(self, main_file_path=None, cohort_file_path=None, std_sample_path=None):
        
        #Setting default encoding as 'utf-8'
        reload(sys)
        sys.setdefaultencoding('utf8')
        
        if(main_file_path == None):
            raise ValueError("Specify the path of the main file in the 'main_file_path' parameter")
        self.df = self.read(main_file_path)
        if(cohort_file_path == None):
            self.cohort_df = None
        else:
            self.cohort_df = self.read(cohort_file_path)
        if(std_sample_path == None):
            self.std_df = None
        else:
            self.std_df = self.read(std_sample_path)
        

    def isFloat(self, value):
        try:
            float(value)
            return True
        except ValueError:
            return False

    def chunkIt(self, seq, num):
        avg = len(seq) / float(num)
        out = []
        last = 0.0

        while last < len(seq):
            out.append(seq[int(last):int(last + avg)])
            last += avg

        return out

    def check_for_sample(self, col,name):
        #Having two conditions of having column size greater than 100 or less than that
        if(col.size >= 100):
            splitted = self.chunkIt(list(col), 100)
            #five selected chunks would be 10, 30, 50, 70, 90
            listt = []
            listt.append(splitted[10])
            listt.append(splitted[30])
            listt.append(splitted[50])
            listt.append(splitted[70])
            listt.append(splitted[90])
            checkedin = []
            #for list 1, checkedin list would be
            checkedin.append(listt[1] + listt[2] + listt[3] + listt[4])
            #for list 2, checkedin list would be
            checkedin.append(listt[0] + listt[2] + listt[3] + listt[4])
            #for list 3, checkedin list would be
            checkedin.append(listt[1] + listt[0] + listt[3] + listt[4])
            #for list 4, checkedin list would be
            checkedin.append(listt[1] + listt[2] + listt[0] + listt[4])
            #for list 5, checkedin list would be
            checkedin.append(listt[1] + listt[2] + listt[3] + listt[0])
            
            max_match = 0
            for i in xrange(5):
                size = len(listt[i])
                temp = [listt[i][0], listt[i][size/2], listt[i][size-1]]
                for word in temp:
                    print("Finding closest matches for word: " + word + " in chunk " + str(i) + " with 0.55 thresold")
                    max_match = max(max_match,len(get_close_matches(word,checkedin[i],n = len(checkedin[i]), cutoff=0.55)))
            
            print("Column " + name + " got closest-matches with " + str((max_match/(len(checkedin[0])))*100) + "%")
            return max_match, (max_match/(len(checkedin[0])))*100
        
        else:
            splitted = self.chunkIt(list(col), 2)
            #2 chunks would be splitted[0] and splitted[1]
            listt = []
            listt.append(splitted[0])
            listt.append(splitted[1])
            checkedin = []
            #for list 1, checkedin list would be
            checkedin.append(listt[1])
            #for list 2, checkedin list would be
            checkedin.append(listt[0])

            max_match = 0
            for i in xrange(2):
                size = len(listt[i])
                temp = [listt[i][0], listt[i][size/2], listt[i][size-1]]
                for word in temp:
                    print("Finding closest matches for word: " + word + " in chunk " + str(i) + " with 0.55 thresold")
                    max_match = max(max_match,len(get_close_matches(word,checkedin[i],n = len(checkedin[i]), cutoff=0.55)))
            
            print("Column " + name + " got closest-matches with " + str((max_match/(len(checkedin[0])))*100) + "%")
            return max_match, (max_match/(len(checkedin[0])))*100


    def check_for_compound(self, col,name):
        if( col.dtype == bool ):
            return 0
        cnt = 0
        idx = 0
        es = Elasticsearch()
        inc = col.size/20
        while (idx < col.size):
            string = col[idx]
            try:
                res = es.search(index="compound", doc_type="my_type", body={"query": {"match": {"name": string}}, "size":0, "terminate_after":1})
                """
                for doc in res['hits']['hits']:
                    print("%s) %s : %s" % (string, doc['_id'], doc['_source']['name']))
                """
                if(res['hits']['total']):
                    cnt = cnt + 1
            except Exception as e:
                print("The type of search keyword {0} is {1}".format(string, type(string)))
                print(e)
            idx = idx + inc
        if(cnt>=15):
            print("Column " + name + " is identified as a compound column!")
            return 1
        else:
            return 0
    
    
    #Method to read the CSV/XLSX files and return the corresponding pandas dataframe
    def read(self, path):
        df = pd.DataFrame()
        if (path.endswith(".csv")):
            try:
                df = pd.read_csv(path, encoding='utf-8', index_col=False)
            except UnicodeDecodeError:
                df = pd.read_csv(path, encoding='ISO-8859-1', index_col=False)
            except Exception as e:
                print("In read(): ")
                print(e)
            
        elif (path.endswith(".xlsx")):
            try:
                df = pd.read_excel(path, encoding='utf-8', index_col=False)
            except UnicodeDecodeError:
                df = pd.read_excel(path, encoding='ISO-8859-1', index_col=False)
            except Exception as e:
                print("In read(): ")
                print(e)
        
        elif (path.endswith(".txt")):
            try:
                df = pd.read_csv(path, encoding='utf-8', index_col=False, delimiter="\t")
            except UnicodeDecodeError:
                df = pd.read_csv(path, encoding='ISO-8859-1', index_col=False, delimiter="\t")
            except Exception as e:
                print("In read(): ")
                print(e)

        else:
            print(path)
            raise ValueError("Unsupported file format")
            return None
        
        return df
    
    def extract_area_columns(self, newdf):

        #Initially mark all the columns as 1 and then eliminate one by one
        new_columns = newdf.columns
        area_mark = [1]*len(new_columns)
        for i in range(len(new_columns)):
            col = newdf[new_columns[i]]
            min_value = col.min(axis=0)
            max_value = col.max(axis=0)
            avg_value = col.mean(axis=0)
            if all( math.isnan(x) for x in col ):
                print("Column no: " + new_columns[i] + " All values are NaN, may be Label or Component no")
                area_mark[i]=0
            elif(avg_value <= 1.0 and avg_value >= 0.0):
                print("Column no: " + new_columns[i] + " Definitely not area, may be Fractional Enrichment or maxQuality or..")
                area_mark[i] = 0
            elif(avg_value >= 0):
                print("Column no: " + new_columns[i] + " can be area column")
            else:
                print("Column no: " + new_columns[i] + " Doesn't seem to be an area column" + str(max_value))
                print(avg_value, min_value, max_value)
                area_mark[i]=0
        print("\nNo of columns passing the range checks( Might be area columns marked but not made sure )")
        print(sum(area_mark))

        #Checking if column names are present or not
        print("\nChecking whether the column names are present")
        float_cnt = 0
        print("Column names which could not be parsed as float or int : \n")
        for i in range(len(area_mark)):
            if(area_mark[i]):
                if(self.isFloat(new_columns[i]) or re.match("^\d+?\.\d+?\.\d+?$", new_columns[i])):
                    float_cnt = float_cnt + 1
                else:
                    print(new_columns[i])

        #print(float_cnt)  
        #Calculation of probable area columns and columns left( to be prompt to the user)
        probable_area_columns = []
        columns_left = []

        if(float_cnt == sum(area_mark)):
            print("Column names are absent\n")
            print("Probable list of columns to be marked as 'Area' are: \n")
            probable_area_columns = [new_columns[i] for i in range(len(new_columns)) if area_mark[i] == 1]
            print(probable_area_columns)
            print(len(probable_area_columns))
            dropped_df = newdf.drop(axis=1,columns = probable_area_columns)
            columns_left = list(dropped_df.columns)
            print("\nNumeric columns except the area columns which are left are :\n")
            print(columns_left)

        else:
            possible_area_columns = list(compress(new_columns, area_mark))
            print("Column names are present\n ")
            #finding close matching words for the column names
            max_matched_5 = 0
            max_matched_4 = 0
            probable_area_columns_5 = []
            probable_area_columns_4 = []
            for i in range(len(area_mark)):
                if(area_mark[i]):
                    word = new_columns[i]
                    print("Columns similar to column '"+new_columns[i]+"' are:")
                    matched_5 = get_close_matches(word,possible_area_columns,n = sum(area_mark), cutoff=0.55)
                    matched_4 = get_close_matches(word,possible_area_columns,n = sum(area_mark), cutoff=0.40)
                    print(matched_5)
                    print("Total matched words: " + str(len(matched_5)))
                    if(max_matched_5 < len(matched_5)):
                        max_matched_5 = len(matched_5)
                        probable_area_columns_5 = matched_5
                    if(max_matched_4 < len(matched_4)):
                        max_matched_4 = len(matched_4)
                        probable_area_columns_4 = matched_4
                    print("\n\n")
            
            columns_left = [] 
            print("Probable list of area columns (with thresold .55) are: \n")
            print(probable_area_columns_5)
            print(len(probable_area_columns_5))

            print("Probable list of area columns (with thresold .40) are: \n")
            print(probable_area_columns_4)
            print(len(probable_area_columns_4))
            
            diff = len(set(probable_area_columns_4) - set(probable_area_columns_5))
            if((diff/len(probable_area_columns_5))*100 <= 10):
                probable_area_columns = probable_area_columns_4
            else:
                probable_area_columns = probable_area_columns_5
            
            print("Final list of probable area columns is: ")
            print(probable_area_columns)

            #Dropping columns which are marked as probable area columns from the dataset
            dropped_df = newdf.drop(axis=1,columns = probable_area_columns)
            columns_left = list(dropped_df.columns)
            
            print("\nNumeric columns except the area columns which are left are :\n")
            print(columns_left)

            #Some more checks to be applied on both the lists ( Probable_area_columns and columns_left)

            #Checking for keywords like 
            #peakIntensity or Intensity or Area
            extracted_list = [s for s in columns_left if ("peakIntensity" in s or "Intensity" in s or "Area" in s or "peakArea" in s or "peakAreaCorrected" in s or "peakAreaTop" in s or "peakAreaTopCorrected" in s)]
            columns_left = list(set(columns_left) - set(extracted_list))
            probable_area_columns = list(set(probable_area_columns).union(set(extracted_list)))

            blacklist_area_columns = ["NA Corrected", "NA Corrected with zero", "Pool_total"]
            temp = [s for s in probable_area_columns if( s not in blacklist_area_columns )]
            columns_left = list(set(columns_left).union(set(probable_area_columns) - set(temp)))
            probable_area_columns = temp

            return probable_area_columns, columns_left

    #Method for checking sample columns if the file is in long format
    def extract_sample_columns(self, non_num_df):
        
        non_num_columns = non_num_df.columns
        sample_mark = [0]*len(non_num_columns)
        probable_sample = []
        percent = [0.0]*len(non_num_columns)
        print("\nChecking for sample columns..\n")

        for i in range(len(non_num_columns)):
                col = non_num_df[non_num_columns[i]]
                if( col.dtype == bool ):
                    print("Column " + non_num_columns[i] +" is bool column")
                    continue
                elif(sum(list(col.str.contains(r"\d+\.\d+\s/\s\d+\.\d+", case = False, regex = True)))):
                    print("Column " + non_num_columns[i] + " contains a pattern which doesn't match that of a sample name")
                    continue
                elif(sum(list(col.str.contains(r"[\w\-]+\s\d*/\d*", case = False, regex = True)))):
                    print("Column " + non_num_columns[i] + " contains a pattern which match that of a compound name")
                    continue
                elif(len(str(col[0])) <= 5):
                    print("Column " + non_num_columns[i] +" contains values whose string length is less than 5")
                    continue
                print("\nChecking the column: " + non_num_columns[i])
                sample_mark[i], percent[i] = self.check_for_sample(col,non_num_columns[i])

        m = max(sample_mark)
        sample_indices = [i for i, j in enumerate(sample_mark) if j == m]
        for i in xrange(len(sample_mark)):
            if(sample_mark[i] == m and sample_mark[i] != 0):
                sample_mark[i] = 1
            else:
                sample_mark[i] = 0
        for i in xrange(len(sample_indices)):
            if(percent[sample_indices[i]] != 0.0):
                probable_sample.append(non_num_columns[sample_indices[i]])

        return probable_sample, sample_mark

    def extract_compound_columns(self, non_num_df,sample_mark):
        
        #First method: Using regex
        non_num_columns = non_num_df.columns
        compound_mark = [0]*len(non_num_columns)
        print("\nChecking for compound columns..\n")
        for i in range(len(non_num_columns)):
            if(sample_mark[i]):
                continue
            col = non_num_df[non_num_columns[i]]
            if( col.dtype == bool or col.dtype in NUMERICS ):
                continue
            cnt = 0
            print("\nChecking for column: " + non_num_columns[i])
            for string in col:
                try:
                    if( re.match(r"[\w\-]+\s\d*/\d*",string.encode('utf-8')) ):
                        cnt = cnt + 1
                except AttributeError as e:
                    print("Type of {0} is interpreted by the program as {1}".format(string,type(string)))
                    print(e)
                except Exception as e:
                    print(e)
                if(cnt >= col.size/2):
                    compound_mark[i] = 1
                    print("Column " + non_num_columns[i] + " is identified as a compound column!")
                    break

        '''
        #Second Method: Using HMDB database indexed on elastic search cluster

        for i in range(len(non_num_columns)):
            if( compound_mark[i] == 0 and sample_mark[i] == 0 ):
                col = non_num_df[non_num_columns[i]]
                cnt = 0
                print("\nChecking for column: " + non_num_columns[i] + " again")
                compound_mark[i] = self.check_for_compound(col,non_num_columns[i])
        
        
        probable_comp = list(compress(non_num_columns, compound_mark))
        return probable_comp, compound_mark 
        '''

    #Method for analyzing the input file and returning probable area, sample and compound columns 
    def analyze(self):
        
        #1. Selecting numeric columns in dataframe
        print("Total columns present in the dataset: \n")
        print(self.df.columns)
        newdf = self.df.select_dtypes(include=NUMERICS)
        new_columns = list(newdf.columns)
        print("Total numeric columns present in dataset: \n")
        print(new_columns)
        print("Total Columns: " + str(len(new_columns)))

        #2. Extracting the area columns
        self.probable_area_columns, columns_left = self.extract_area_columns(newdf)   
        print("\nRefined list of probable Area columns is:\n")
        print(self.probable_area_columns)
        print("\n")
        print("Refined list of Columns left besides area columns:\n")
        print(columns_left)

        #3. Selecting non-numeric columns in dataframe
        non_num_df = self.df.select_dtypes(exclude=NUMERICS)
        print("\nNon-numeric columns present are:\n")
        non_num_columns = non_num_df.columns
        print(non_num_columns)

        #4. Extracting the sample columns    
        self.probable_sample, sample_mark = self.extract_sample_columns(non_num_df)
        print("\nSample columns predicted with o.55 thresold with maximum closest matches are :\n")
        print(self.probable_sample)

        '''
        #5. Extracting the compound columns 
        self.probable_comp, compound_mark = self.extract_compound_columns(non_num_df, sample_mark)
        print("Probable list of compound columns can be:\n")
        print(self.probable_comp)  
        '''

        #6.Getting the format of the file
        self.form = ""
        if (len(self.probable_sample) < 1):
            self.form = "wide"
        else:
            self.form = "long"
        
        return self.df, self.probable_area_columns, self.probable_sample, self.form

    #Returns dataframe corresponding to cohort file
    def get_cohort_df(self):
        return self.cohort_df
    
    #Returns dataframe corresponding to metadata file
    def get_std_df(self):
        return self.std_df

    #Returns dataframe  corresponding to main file
    def get_df(self):
        return self.df

if __name__ == "__main__":
    print("Executing this module as main program!!")